@@ -12,14 +12,15 @@ use Carp;
 use IPC::Open3;
 use File::Spec;
 use POSIX qw(:sys_wait_h);
+use File::Temp qw(tempfile);
 
 use vars qw(@ISA @EXPORT @EXPORT_OK $VERSION %pids %dead_pids
-	    %stdout_handles %stderr_handles $debug);
+	    %stdout_handles %stderr_handles $debug @events);
 
 @ISA       = qw(Exporter);
 @EXPORT    = qw();
 @EXPORT_OK = qw(start_job watch_jobs);
-$VERSION   = 0.04;
+$VERSION   = 0.05;
 
 sub new_handle ();
 sub is_our_handle ($);
@@ -50,6 +51,7 @@ sub start_job {
     my($stdin_handle, $stdout_handle, $stderr_handle);
     my($capture_stdout, $capture_stderr);
     my($pid);
+    my(@to_close);
 
     if ($_[0] eq __PACKAGE__) {
 	# Should we flame at the user for calling this function as a
@@ -108,6 +110,7 @@ sub start_job {
 	    or croak "Couldn't reopen fhandle $params{'stdout_handle'}: $!";
 	warn("Stdout file descriptor (from handle) is " .
 	     fileno($stdout_handle) . "\n") if ($debug);
+	push(@to_close, $stdout_handle);
 	$stdout_handle = '>&' . $stdout_handle;
 	delete $params{'stdout_handle'};
     }
@@ -130,6 +133,7 @@ sub start_job {
 	no strict 'refs';
 	open($stderr_handle, '>&' . fileno($params{'stderr_handle'}))
 	    or croak "Couldn't reopen fhandle $params{'stderr_handle'}: $!";
+	push(@to_close, $stderr_handle);
 	$stderr_handle = '>&' . $stderr_handle;
 	delete $params{'stderr_handle'};
     }
@@ -176,6 +180,8 @@ sub start_job {
 	     "\n") if ($debug);
     }
 
+    map(close($_), @to_close);
+
     return $pid;
 }
 
@@ -195,9 +201,7 @@ sub watch_jobs {
     my(%stdout_to_pid, %stderr_to_pid);
     my($nfound, $waiting);
     my($pid, $type);
-    my($size) = 1024;
     my($ret);
-    my($data) = '';
     my($rbits, $ebits);
     my(@pids) = keys %pids;
     my(@dead_pids) = keys %dead_pids;
@@ -208,8 +212,7 @@ sub watch_jobs {
     }
 
     foreach my $pid (@pids, @dead_pids) {
-	my($capturing) = $stdout_handles{$pid} || $stderr_handles{$pid};
-	if (! ($capturing || $dead_pids{$pid})) {
+	if (! ($stdout_handles{$pid} || $stderr_handles{$pid})) {
 	    push(@pid_only, $pid);
 	}
 	else {
@@ -224,6 +227,10 @@ sub watch_jobs {
 
     if (! @dead_pids) {
 	if (@pid_only == @pids) {
+	    if (@events) {
+		return(@{shift @events});
+	    }
+
 	    warn "Waiting for @pids to exit\n" if ($debug);
 	    $pid = wait;
 	    if ($pid == -1) {
@@ -244,76 +251,90 @@ sub watch_jobs {
 	    if ($stdout_handles{$pid} || $stderr_handles{$pid}) {
 		$dead_pids{$pid}++;
 	    }
-	    return($pid, 'EXIT', $?);
+	    push(@events, [$pid, 'EXIT', $?]);
 	}
     }
 
-    $rbits = '';
-    map((vec($rbits, fileno($_), 1) = 1),
-	keys %stdout_to_pid,
-	keys %stderr_to_pid);
-
-    warn "Calling select() to wait for output from jobs\n" if ($debug);
-
-    $nfound = select($rbits, undef, $ebits = $rbits, undef);
-
-    die(__PACKAGE__ . ": Internal error: Select unexpectedly " .
-	"returned no pending data") if (! $nfound);
-
-    ($waiting) = grep(vec($rbits, fileno($_), 1) || vec($ebits, fileno($_), 1),
-		      keys %stdout_to_pid, keys %stderr_to_pid);
-
-    die(__PACKAGE__ . ": Internal error: Select returned " .
-	"an unexpected bit") if (! $waiting);
-
-    if (($pid = $stdout_to_pid{$waiting})) {
-	$type = 'STDOUT';
-    }
-    elsif (($pid = $stderr_to_pid{$waiting})) {
-	$type = 'STDERR';
-    }
-    else {
-	die(__PACKAGE__ . ": Internal error: unexpected file handle");
-    }
-
-    # Read all pending data
-    do {
+    while (1) {
 	$rbits = '';
-	vec($rbits, fileno($waiting), 1) = 1;
-	$ret = sysread($waiting, $data, $size, length($data));
-	$size *= 2;
-	$nfound = select($rbits, undef, $ebits = $rbits, 0);
-    } while ($ret && $nfound);
+	map((vec($rbits, fileno($_), 1) = 1),
+	    keys %stdout_to_pid,
+	    keys %stderr_to_pid);
 
+	warn "Calling select() to wait for output from jobs\n" if ($debug);
 
-    if ($data) {
-	warn "Got $type output ($data) from $pid\n" if ($debug);
-    }
-    else {
-	# Got EOF
-	warn "Got $type EOF from $pid\n" if ($debug);
-	close($waiting);
-	delete $stdout_handles{$pid} if ($type eq 'STDOUT');
-	delete $stderr_handles{$pid} if ($type eq 'STDERR');
-	if ($dead_pids{$pid} &&
-	    ! ($stderr_handles{$pid} || $stdout_handles{$pid})) {
-	    delete $dead_pids{$pid};
+	$nfound = select($rbits, undef, $ebits = $rbits, @events ? 0 : undef);
+
+	if (@events && ! $nfound) {
+	    last;
 	}
+
+	die(__PACKAGE__ . ": Internal error: Select unexpectedly " .
+	    "returned no pending data") if (! $nfound);
+
+	($waiting) = grep(vec($rbits, fileno($_), 1) || vec($ebits, fileno($_), 1),
+			  keys %stdout_to_pid, keys %stderr_to_pid);
+
+	die(__PACKAGE__ . ": Internal error: Select returned " .
+	    "an unexpected bit") if (! $waiting);
+
+	if (($pid = $stdout_to_pid{$waiting})) {
+	    $type = 'STDOUT';
+	}
+	elsif (($pid = $stderr_to_pid{$waiting})) {
+	    $type = 'STDERR';
+	}
+	else {
+	    die(__PACKAGE__ . ": Internal error: unexpected file handle");
+	}
+
+	# Read all pending data
+	my($nfound2);
+	my($size) = 1024;
+	my($data) = '';
+	do {
+	    $rbits = '';
+	    vec($rbits, fileno($waiting), 1) = 1;
+	    $ret = sysread($waiting, $data, $size, length($data));
+	    $size *= 2;
+	    $nfound2 = select($rbits, undef, $ebits = $rbits, 0);
+	} while ($ret && $nfound2);
+
+	if ($data) {
+	    warn "Got $type output ($data) from $pid\n" if ($debug);
+	}
+	else {
+	    # Got EOF
+	    warn "Got $type EOF from $pid\n" if ($debug);
+	    close($waiting);
+	    if ($type eq 'STDOUT') {
+		delete $stdout_to_pid{$stdout_handles{$pid}};
+		delete $stdout_handles{$pid};
+	    }
+	    if ($type eq 'STDERR') {
+		delete $stderr_to_pid{$stderr_handles{$pid}};
+		delete $stderr_handles{$pid};
+	    }
+	    if ($dead_pids{$pid} &&
+		! ($stderr_handles{$pid} || $stdout_handles{$pid})) {
+		delete $dead_pids{$pid};
+	    }
+	}
+
+	push(@events, [$pid, $type, $data]);
     }
 
-    return($pid, $type, $data);
+    return(@{shift @events});
 }
 
 # If the argument is true, operate silently.
 
 sub test ($) {
-    my($outfile) = sprintf('/tmp/%s.test.%d', __PACKAGE__, $$);
+    my($fh, $outfile) = tempfile(__PACKAGE__ . ".testXXXX");
     my(@tests);
 
     $debug = ! (@_ && $_[0]);
 
-    $outfile = undef if (! open(OUTFILE, ">$outfile"));
-
     my(@commands) = (
 		     [ 'true', ],
 		     [ 'sleep', '1' ],
@@ -324,7 +345,7 @@ sub test ($) {
     @tests = ({}, {stdout_capture=>1}, {stderr_capture=>1},
 	      {stdout_capture=>1,stderr_capture=>1});
     if ($outfile) {
-	push(@tests, {stdout_handle=>"OUTFILE"});
+	push(@tests, {stdout_handle=>$fh});
     }
 
     if (! $debug) {
@@ -377,7 +398,7 @@ sub test ($) {
     if ($outfile) {
 	my($text);
 
-	close(OUTFILE);
+	close($fh);
 
 	{
 	    local($/) = undef;
@@ -482,11 +503,6 @@ Note that it is possible to receive STDOUT or STDERR events from a
 process after its EXIT event, i.e., you may receive an EXIT event
 before you've read all of a process's output.
 
-If you mix some jobs for which you are capturing stdout and/or stderr
-with some jobs for which you are not, watch_jobs() may sometimes not
-notice that a job whose output isn't being captured has exited until a
-job whose output B<is> being captured exits or generates output.
-
 =head1 AUTHOR
 
 Jonathan Kamens E<lt>jik@kamens.brookline.ma.usE<gt>
